/*:
 * @target MZ
 * @plugindesc Simple Day/Night Cycle - Street Lighting
 * @author Your Name
 * @help Adds a simple day/night cycle with lighting on a specific map.
 *
 * Map Note: <StreetDayNight> - Tag to enable day/night cycle on this map.
 *
 * @param Minutes Per Step
 * @desc How many minutes pass for each map update.
 * @default 5
 *
 * @param Day Start Hour
 * @desc The hour when day starts (0-23).
 * @default 6
 *
 * @param Night Start Hour
 * @desc The hour when night starts (0-23).
 * @default 18
 */

(function() {

    const parameters = PluginManager.parameters('StreetDayNightCycle');
    const minutesPerStep = parseInt(parameters['Minutes Per Step'] || 5);
    const dayStartHour = parseInt(parameters['Day Start Hour'] || 6);
    const nightStartHour = parseInt(parameters['Night Start Hour'] || 18);
  
    const dayTone = [0, 0, 0, 0]; // Дневная тональность (без изменений)
    const nightTone = [-100, -100, -100, 0]; // Ночная тональность (темнее)
  
    let totalMinutes = 0;
  
    function isNight() {
      const currentHour = Math.floor((totalMinutes / 60) % 24);
      return currentHour >= nightStartHour || currentHour < dayStartHour;
    }
  
    // ** Game_Map **
    const _Game_Map_update = Game_Map.prototype.update;
    Game_Map.prototype.update = function(sceneActive) {
      _Game_Map_update.call(this, sceneActive);
  
      if ($dataMap && $dataMap.note && $dataMap.note.includes("<StreetDayNight>")) {
        totalMinutes += minutesPerStep;
  
        let tone = dayTone;
        if (isNight()) {
          tone = nightTone;
        }
  
        $gameScreen.startTint(tone, 60);
      }
    };
  
  })();