/*:
 * @target MZ
 * @plugindesc Simple Day/Night Cycle with Lighting (Event Triggered) - DEBUG v3
 * @author Your Name
 * @help This plugin adds a basic day/night cycle and changes map lighting, triggered by an event.
 *
 * Map Note: <DayNightCycleMap> - Tag to enable time flow on the map.
 * Event Note: <DayNightLight> - Add to an event to enable dynamic lighting.
 *
 * @param Minutes Per Step
 * @default 1
 *
 * @param Day Start Hour
 * @default 6
 *
 * @param Night Start Hour
 * @default 18
 */

(function() {

  const parameters = PluginManager.parameters('SimpleDayNightCycle');
  const minutesPerStep = parseInt(parameters['Minutes Per Step'] || 5);
  const dayStartHour = parseInt(parameters['Day Start Hour'] || 6);
  const nightStartHour = parseInt(parameters['Night Start Hour'] || 18);

  const dayTone = [0, 0, 0, 0];
  const nightTone = [-200, -200, -100, 0];

  let totalMinutes = 0;

  function isNight() {
    const currentHour = Math.floor((totalMinutes / 60) % 24);
    console.log(`[isNight] currentHour: ${currentHour}, nightStartHour: ${nightStartHour}, dayStartHour: ${dayStartHour}`);
    return currentHour >= nightStartHour || currentHour < dayStartHour;
  }

  // ** Game_Map **
  const _Game_Map_update = Game_Map.prototype.update;
  Game_Map.prototype.update = function(sceneActive) {
    _Game_Map_update.call(this, sceneActive);
    if (this.isDayNightCycleMap() && this.hasDayNightLightEvent()) {
      totalMinutes += minutesPerStep;
      console.log(`[update] totalMinutes: ${totalMinutes}`);

      let tone = dayTone; // По умолчанию - дневная тональность
      if (isNight()) {
        tone = nightTone;
        console.log("[update] It's night!");
      } else {
        console.log("[update] It's day!");
      }

      console.log(`[update] Setting tone to: ${tone}`);
      $gameScreen.startTint(tone, 60);
    }
  };

  Game_Map.prototype.isDayNightCycleMap = function() {
    return $dataMap && $dataMap.note && $dataMap.note.includes("<DayNightCycleMap>");
  };

  Game_Map.prototype.hasDayNightLightEvent = function() {
    if (!$dataMap) return false;
    for (const event of $dataMap.events) {
      if (event && event.note && event.note.includes("<DayNightLight>")) {
        return true;
      }
    }
    return false;
  };

})();